#!/usr/bin/env python3
import os
import sys
import re
import subprocess
import requests
import logging
import stat
from typing import List, Tuple, Optional

# Configure logging
logging.basicConfig(level=logging.INFO, format='[%(levelname)s] %(message)s')

# OpenAI API configuration
OPENAI_API_URL = "https://api.openai.com/v1/chat/completions"
DEFAULT_MODEL = "gpt-4o"
DEFAULT_MAX_TOKENS = 500
API_TIMEOUT = 20  # seconds
SUBPROCESS_TIMEOUT = 5  # seconds

SYSTEM_PROMPT = (
    "You are a senior developer tasked with reviewing code changes provided as an annotated git diff. "
    "Each added or context line has been annotated with its corresponding line number from the updated file. "
    "Your job is to identify any syntax errors, potential bugs, security concerns, and deviations from best practices, "
    "ignoring any issues related to code style, linting, or missing newlines at the end of files. "
    "When referencing a specific issue, use the annotated line numbers provided. "
    "Your review must strictly adhere to the following format and rules:\n\n"
    "1. **Blocking Issues Section:**\n"
    "   - Begin with a single line that contains exactly:\n\n"
    "     Blocking issues:\n\n"
    "2. **Issue Reporting:**\n"
    "   - For each blocking issue found, output a bullet point on its own line following this exact format:\n\n"
    "     ❌ File: <filename>, Line: <line number>: <description of the issue>.\n\n"
    "   - **Details:**\n"
    "     - `<filename>` is the name of the file where the issue was detected.\n"
    "     - `<line number>` must correspond to the line number in the updated file as provided in the annotation. If no line number is available, write `N/A`.\n"
    "     - `<description of the issue>` is a concise statement explaining the problem.\n\n"
    "3. **Separation of Multiple Issues:**\n"
    "   - If there are multiple blocking issues, separate each bullet point with a line that contains exactly:\n\n"
    "     ---\n\n"
    "4. **Final Verdict:**\n"
    "   - After listing all blocking issues, output exactly one final line with either:\n\n"
    "     Final verdict: COMMIT PASS\n\n"
    "     or\n\n"
    "     Final verdict: COMMIT ABORTED\n\n"
    "   - This final line must be the very last line of your output.\n\n"
    "Follow these instructions exactly."
)

def run_subprocess(command: List[str]) -> str:
    try:
        result = subprocess.run(
            command,
            check=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
            timeout=SUBPROCESS_TIMEOUT
        )
        return result.stdout
    except subprocess.TimeoutExpired:
        logging.error(f"Command {' '.join(command)} timed out.")
        sys.exit(1)
    except subprocess.CalledProcessError as e:
        logging.error(f"Command {' '.join(command)} failed: {e.stderr.strip()}")
        sys.exit(1)

def annotate_diff_with_line_numbers(diff: str) -> str:
    """
    Annotates each context (" ") and added ("+") line in the diff with its corresponding line number
    from the updated file. This makes it easier for downstream processing (and the AI) to reference
    the exact line numbers.
    """
    annotated_lines = []
    current_file = None
    new_line_number = None

    for line in diff.splitlines():
        if line.startswith("diff --git"):
            parts = line.split()
            if len(parts) >= 4:
                # The new file is typically prefixed with "b/"
                current_file = parts[3][2:]
            annotated_lines.append(line)
        elif line.startswith("+++ "):
            annotated_lines.append(line)
        elif line.startswith("@@"):
            m = re.search(r'\+(\d+)(?:,(\d+))?', line)
            if m:
                new_line_number = int(m.group(1))
            else:
                new_line_number = None
            annotated_lines.append(line)
        elif line.startswith(" "):
            if new_line_number is not None and current_file is not None:
                annotated_line = f"[{current_file}:{new_line_number}] {line}"
                new_line_number += 1
            else:
                annotated_line = line
            annotated_lines.append(annotated_line)
        elif line.startswith("+"):
            if new_line_number is not None and current_file is not None:
                annotated_line = f"[{current_file}:{new_line_number}] {line}"
                new_line_number += 1
            else:
                annotated_line = line
            annotated_lines.append(annotated_line)
        else:
            annotated_lines.append(line)
    return "\n".join(annotated_lines)

def get_diff_output() -> str:
    raw_diff = run_subprocess(["git", "diff", "--cached"])
    if raw_diff:
        return annotate_diff_with_line_numbers(raw_diff)
    return raw_diff

def call_openai_api(session: requests.Session, api_key: str, diff_output: str) -> Tuple[str, Optional[int]]:
    payload = {
        "model": DEFAULT_MODEL,
        "messages": [
            {"role": "system", "content": SYSTEM_PROMPT},
            {"role": "user", "content": diff_output}
        ],
        "max_tokens": DEFAULT_MAX_TOKENS
    }
    headers = {
        "Authorization": f"Bearer {api_key}",
        "Content-Type": "application/json"
    }
    try:
        response = session.post(
            OPENAI_API_URL,
            headers=headers,
            json=payload,
            timeout=API_TIMEOUT
        )
        response.raise_for_status()
    except requests.RequestException as e:
        logging.error(f"Error calling OpenAI API: {e}")
        sys.exit(1)

    try:
        json_response = response.json()
        message = json_response["choices"][0]["message"]["content"].strip()
        tokens_used = json_response.get("usage", {}).get("total_tokens")
    except (KeyError, IndexError):
        logging.error("Unexpected response structure from OpenAI API.")
        sys.exit(1)
    
    return message, tokens_used

def should_ignore_hook() -> bool:
    """
    Determines if the hook should be skipped based on the presence
    of the environment variable SKIP_AI_REVIEW set to '1'.
    """
    return os.getenv("SKIP_AI_REVIEW") == "1"

def main():
    """
    Main entry point for the AI review hook.
    This function is called when running `openai-git-hook` (either directly or via a Git hook).
    """
    if should_ignore_hook():
        logging.info("AI review skipped due to SKIP_AI_REVIEW environment variable.")
        sys.exit(0)

    api_key = os.getenv("OPENAI_API_KEY")
    if not api_key:
        logging.error("OpenAI API key not set. Please set OPENAI_API_KEY environment variable.")
        sys.exit(1)

    diff_output = get_diff_output()
    if not diff_output:
        logging.info("No diff output available. Skipping review.")
        sys.exit(0)

    logging.info("Reviewing staged changes with OpenAI API...")
    with requests.Session() as session:
        review, tokens_used = call_openai_api(session, api_key, diff_output)
        print(review)
        if tokens_used is not None:
            print(f"\nTokens used: {tokens_used}")

    if "Final verdict: COMMIT ABORTED" in review:
        sys.exit(1)
    elif "Final verdict: COMMIT PASS" in review:
        sys.exit(0)
    else:
        logging.error("No clear verdict provided. Commit aborted by default.")
        sys.exit(1)

def install_hook():
    """
    Installs (or updates) the pre-commit hook in the current Git repository.
    This function writes a pre-commit script that calls the installed `openai-git-hook` command.
    """
    git_hooks_dir = os.path.join(os.getcwd(), ".git", "hooks")
    if not os.path.isdir(git_hooks_dir):
        print("[ERROR] No .git/hooks directory found. Are you in a Git repository?")
        sys.exit(1)

    pre_commit_path = os.path.join(git_hooks_dir, "pre-commit")

    script_contents = """#!/bin/bash
# Generated by openai-git-hook
openai-git-hook
"""

    try:
        with open(pre_commit_path, "w") as f:
            f.write(script_contents)
        current_mode = os.stat(pre_commit_path).st_mode
        os.chmod(pre_commit_path, current_mode | stat.S_IEXEC)
        print(f"[INFO] Installed pre-commit hook at {pre_commit_path}")
    except OSError as e:
        print(f"[ERROR] Unable to write pre-commit hook: {e}")
        sys.exit(1)

if __name__ == '__main__':
    main()
